import { it, expect, describe, beforeEach, afterEach, vi } from 'vitest'
import { mount } from '@vue/test-utils'
import { Quasar } from 'quasar'
import { i18n } from '@/plugins/i18n'
import HeaderMenu from '@/components/menu/notifications/HeaderMenu.vue'
import Notification from '@/components/menu/notifications/SingleItem.vue'
import { useDialogStore, NOTIFICATION_INTERVAL } from '@/stores/dialog'
import { createPinia, setActivePinia } from 'pinia'
import { nextTick } from 'vue'
import type { AppNotification } from '../../../../types/notification'
import type { TestWrapper } from '../../types'


const newNotif = (id: number): AppNotification => ({
  id,
  title: 'title',
  timesince: 'almost',
  unread: true
})


describe('HeaderMenu', () => {
  setActivePinia(createPinia())
  const store = useDialogStore()

  let wrapper: TestWrapper<{ panelIsOpen: boolean }>

  const mountComponent = () => {
    wrapper = mount(HeaderMenu, {global: {plugins: [i18n, Quasar]}})
  }

  beforeEach(()=> {
    store.notify = {
      previousPage: null,
      nextPage: 2,
      count: 7,
      notifications: [1,2,3].map(n=>newNotif(n))
    }
    store.getNotifications = vi.fn()
    mountComponent()
  })
  afterEach(()=> {
    vi.restoreAllMocks()
    if (wrapper) wrapper.unmount()
  })
  describe('closed menu', () => {
    it('should display unread count', () => {
      const buttons = wrapper.findAll('button')
      expect(buttons).toHaveLength(1)
      expect(buttons[0].text()).toContain(7)
    })
    it('should open a list of notifications', async () => {
      expect(wrapper.findAllComponents(Notification)).toHaveLength(0)
      const button = wrapper.find('button')
      await button.trigger('click')
      await nextTick()
      expect(wrapper.findAllComponents(Notification)).toHaveLength(3)
    })
    // menu is teleported elsewhere in DOM
    // it('should display action buttons', async () => {
    //   const button = wrapper.find('button')
    //   await button.trigger('click')
    //   await nextTick()
    //   expect(wrapper.findAll('button')).toHaveLength(4)
    // })
  })
  describe('notification refresh', () => {
    const inSecondInterval = Math.ceil(NOTIFICATION_INTERVAL*1.5)
    it('should refresh after time interval if panel is closed', async () => {
      const testTimeout = () => new Promise(resolve => {
        setTimeout(resolve, inSecondInterval)
      }).then(() => {
        expect(store.getNotifications).toHaveBeenCalledTimes(2)
      })
      await testTimeout()
    })
    it('should not refresh after time interval if panel is open', async () => {
      wrapper.vm.panelIsOpen = true
      const testTimeout = () => new Promise(resolve => {
        setTimeout(resolve, inSecondInterval)
      }).then(() => {
        expect(store.getNotifications).toHaveBeenCalledOnce()
      })
      await testTimeout()
    })
    it('should not refresh if component is destroyed', async () => {
      setTimeout(()=>wrapper.unmount(), NOTIFICATION_INTERVAL/2)
      const testTimeout = () => new Promise(resolve => {
        setTimeout(resolve, inSecondInterval)
      }).then(() => {
        expect(store.getNotifications).toHaveBeenCalledTimes(1)
      })
      await testTimeout()
    })
  })
})
