import { describe, it, expect, beforeEach } from 'vitest'

import { useRules } from '@/composables/useRules'
import { t } from '@/plugins/i18n'
import { ref } from 'vue'
import type { ComputedRef } from 'vue'

type Rule = (val: any) => true | string


describe('useRules', () => {
  describe('with no passwords ref argument', () => {
    it('should return two computed if no password ref is given', () => {
      const {requiredField, passwordField} = useRules()
      expect(requiredField).toBeTypeOf('object')
      expect(passwordField).toBeTypeOf('object')
    })
    describe('requiredField', () => {
      const requiredError = t('form.rule.required')

      let requiredRule: Rule
      beforeEach(() => {
        const { requiredField } = useRules()
        requiredRule = requiredField.value[0]
      })
      it('should return an error message if text field value is undefined', () => {
        expect(requiredRule(undefined)).toEqual(requiredError)
      })
      it('should return an error string if text field is empty', () => {
        expect(requiredRule('')).toEqual(requiredError)
      })
      it('should return true if text field is not empty', () => {
        expect(requiredRule('a')).toEqual(true)
      })
    })
    describe('passwordField', () => {
      const tooShortError = t('form.rule.minPassLength')
      let passwordRules: Rule[], requiredRule: Rule, lengthRule: Rule
      beforeEach(() => {
        const {requiredField, passwordField} = useRules()
        requiredRule = requiredField.value[0]
        passwordRules = passwordField.value
        lengthRule = passwordRules[1]
      })
      it('should contains 2 rules', () => {
        expect(passwordRules).toHaveLength(2)
      })
      it('first rule should be requiredField rule', () => {
        expect(passwordRules[0]).toEqual(requiredRule)
      })
      it('should return an error strings if text field value is undefined', () => {
        expect(lengthRule(undefined)).toEqual(tooShortError)
      })
      it('should return an error string if text field value is too short', () => {
        expect(lengthRule('short')).toEqual(tooShortError)
      })
      it('should return true if text field value is long enough', () => {
        expect(lengthRule('longEnoughPass')).toBe(true)
      })
    })
  })
  describe('with passwords ref argument', () => {
    const passToBeConfirmed = 'toBeConfirmed'
    const passwordsRef = ref({
      newPassword: passToBeConfirmed,
      newPassword2: 'user input'
    })
    it('should return two computed if no password ref is given', () => {
      const {requiredField, passwordField} = useRules(passwordsRef)
      expect(requiredField).toBeTypeOf('object')
      expect(passwordField).toBeTypeOf('object')
    })
    describe('passwordMatch', () => {
      let matchRules: Rule[], passwordRules: Rule[]
      let matchRule: Rule
      beforeEach(() => {
        const {passwordField, passwordMatch} = useRules(passwordsRef)
        passwordRules = passwordField.value
        matchRules = (passwordMatch as ComputedRef).value
        matchRule = matchRules[2]
      })
      it('should contains 3 rules', () => {
        expect(matchRules).toHaveLength(3)
      })
      it('first rules should be passwordField rules', () => {
        expect(matchRules.slice(0,2)).toEqual(passwordRules)
      })
      it('should return an error string if text field value is too short', () => {
        expect(matchRule('notMatchingPass')).toEqual(t('form.rule.passMatch'))
      })
      it('should return true if text field value is long enough', () => {
        expect(matchRule(passToBeConfirmed)).toBe(true)
      })
    })
  })
})
