import { describe, it, expect, beforeEach, afterEach, vi } from 'vitest'
import { createPinia, setActivePinia } from 'pinia'

import { useReferentialStore } from '@/stores/referential'
import { getLevelNodes, getReferential, postReferential } from '@/services/referentialsService'

import type { NewReferential, Node, NodeBase, Referential } from '#/referential'
import type { SpyInstance } from 'vitest'


vi.mock('@/services/alertService', () => ({
      notifyError: vi.fn(),
      notifySuccess: vi.fn()
    })
)


vi.mock('@/services/referentialsService', () => ({
      getLevelNodes: vi.fn(),
      // getProjectReferentials: vi.fn(),
      getReferential: vi.fn(),
      // patchReferential: vi.fn(),
      // patchScope: vi.fn(),
      postReferential: vi.fn(),
    })
)

const levelNodesMock = vi.mocked(getLevelNodes, true)
const postReferentialMock = vi.mocked(postReferential, true)
const getReferentialMock = vi.mocked(getReferential, true)

const firstLevel = {
  id: 'dadcefe6-c5f4-4038-9b62-1c8316835775',
  name: 'niveau 1'
}

const secondLevel = {
  id: 'cdb58bb4-bf24-42b0-a2dd-96af83349b24',
  name: 'niveau 2'
}

const firstRootId = 'b0e54e52-c95d-435e-910c-d0190ef2df9c'

const referentialId = 'a8b14a8e-2c44-4423-afcf-62dd70de2c9a'

const initDefinition = () => ({
  levels: [
    firstLevel,
    secondLevel,
    {
      id: 'a4504b31-a94c-4070-8311-4082da4f42e1',
      name: 'niveau 3'
    }
  ],
  roots: [
    'b0e54e52-c95d-435e-910c-d0190ef2df9c',
    '8f3973b5-ab1f-4ebc-8e94-71984d723966'
  ],
  nodes: {
    [firstRootId]: {
      id: firstRootId,
      name: 'racine1',
      level: 0,
      children: [
        '4ef38584-63e9-4b60-a666-e9d0288b84c5',
        'bf874f02-d051-44f0-8f5e-5e0d93fa83a3',
        'fbce0804-6c81-43d1-bea8-f51ceae751e6'
      ],
      parents: []
    },
    '4ef38584-63e9-4b60-a666-e9d0288b84c5': {
      id: '4ef38584-63e9-4b60-a666-e9d0288b84c5',
      name: 'nœud11',
      level: 1,
      children: [
        '467e6bd7-671d-44f0-990b-72488a8725fd',
        'c6af0c20-3399-494c-90a4-dbeff04ace2d'
      ],
      parents: ['b0e54e52-c95d-435e-910c-d0190ef2df9c']
    },
    '467e6bd7-671d-44f0-990b-72488a8725fd': {
      id: '467e6bd7-671d-44f0-990b-72488a8725fd',
      name: 'nœud111',
      level: 2,
      parents: ['4ef38584-63e9-4b60-a666-e9d0288b84c5']
    },
    'c6af0c20-3399-494c-90a4-dbeff04ace2d': {
      id: 'c6af0c20-3399-494c-90a4-dbeff04ace2d',
      name: 'nœud112',
      level: 2,
      parents: ['4ef38584-63e9-4b60-a666-e9d0288b84c5']
    },
    'bf874f02-d051-44f0-8f5e-5e0d93fa83a3': {
      id: 'bf874f02-d051-44f0-8f5e-5e0d93fa83a3',
      name: 'nœud12',
      level: 1,
      parents: ['b0e54e52-c95d-435e-910c-d0190ef2df9c']
    },
    'fbce0804-6c81-43d1-bea8-f51ceae751e6': {
      id: 'fbce0804-6c81-43d1-bea8-f51ceae751e6',
      name: 'nœud101',
      level: 2,
      parents: ['b0e54e52-c95d-435e-910c-d0190ef2df9c']
    },
    '8f3973b5-ab1f-4ebc-8e94-71984d723966': {
      id: '8f3973b5-ab1f-4ebc-8e94-71984d723966',
      name: 'racine2',
      level: 0,
      children: [
        '7bf767f3-f834-4b86-baf8-e66e2dfdfac8',
        '63c2ff74-1c83-4fea-8e06-e1037e430c4d'
      ],
      parents: []
    },
    '7bf767f3-f834-4b86-baf8-e66e2dfdfac8': {
      id: '7bf767f3-f834-4b86-baf8-e66e2dfdfac8',
      name: 'nœud21',
      level: 1,
      children: ['c8fd76ae-32f4-466a-adb3-a04abd58912e'],
      parents: ['8f3973b5-ab1f-4ebc-8e94-71984d723966']
    },
    'c8fd76ae-32f4-466a-adb3-a04abd58912e': {
      id: 'c8fd76ae-32f4-466a-adb3-a04abd58912e',
      name: 'nœud211',
      level: 2,
      parents: ['7bf767f3-f834-4b86-baf8-e66e2dfdfac8']
    },
    '63c2ff74-1c83-4fea-8e06-e1037e430c4d': {
      id: '63c2ff74-1c83-4fea-8e06-e1037e430c4d',
      name: 'nœud201',
      level: 2,
      parents: ['8f3973b5-ab1f-4ebc-8e94-71984d723966']
    },
  },
})

const initReferential = (): Referential => ({
  id: referentialId,
  project: {id: 'de9f6ed4-eb2a-4b09-b48d-fae091e2cf17', name: 'test'},
  creator: {id: '56139eb4-5845-491e-8a2e-b74c519cf683', name: 'user test'},
  name: 'suppression test',
  description: 'jeu de test',
  scope: 4,
  version: null,
  versionLabel: null,
  currentVersion: true,
  source: 'user test',
  languages: {default: '', available: [], translations: {}},
  definition: initDefinition(),
  lists: [],
  createdAt: '2023-03-17T15:06:50.711241+01:00',
  updatedAt: '2023-03-17T15:06:50.711268+01:00',
  rootVersion: null,
})

describe('store/referential', () => {
  const baseReferential = initReferential()

  const firstRootNode = baseReferential.definition.nodes[baseReferential.definition.roots[0]] as Required<Node>

  setActivePinia(createPinia())
  const store = useReferentialStore()

  beforeEach(() => {
    store.$reset()
  })
  afterEach(() => {
    vi.restoreAllMocks()
  })
  describe('actions', () => {
    describe('referential build', () => {
      beforeEach(() => {
        store.referential = initReferential()
      })
      describe('nodeAdd', () => {
        const nodeLevel = 2
        const nodeName = 'node name'
        let createSpy: SpyInstance<[string, number], Node>
        beforeEach(() => {
          createSpy = vi.spyOn(store, 'nodeCreate')
        })
        it('should create a new node', () => {
          store.nodeAdd(nodeName, nodeLevel)
          expect(createSpy).toHaveBeenCalledOnce()
        })
        it('should add node id to roots if no ancestor is provided', () => {
          store.nodeAdd(nodeName, nodeLevel)
          const nodeId = createSpy.mock.results[0].value.id
          const roots = [...store.referential.definition.roots]
          expect(roots.pop()).toEqual(nodeId)
        })
        it('should add node id as last children of its ancestor if provided', () => {
          const ancestorId = firstRootNode.id
          store.nodeAdd(nodeName, nodeLevel, ancestorId)
          const nodeId = createSpy.mock.results[0].value.id
          expect((store.nodes[ancestorId].children as string[]).pop()).toEqual(nodeId)
        })
      })
      describe('nodeCreate', () => {
        const nodeLevel = 2
        const nodeName = 'node name'
        let createdNode: Node
        beforeEach(() => {
          createdNode = store.nodeCreate(nodeName, nodeLevel)
        })
        it('should return a node with id', () => {
          expect(Object.keys(createdNode).includes('id')).toBe(true)
          expect(typeof createdNode.id).toBe('string')
          expect(createdNode.name).toEqual(nodeName)
          expect(createdNode.level).toEqual(nodeLevel)
        })
        it('should add this node to referential node list', () => {
          expect(store.referential.definition.nodes[createdNode.id]).toStrictEqual(createdNode)
        })
      })
      describe('nodeDelete', () => {
        it('should remove targeted node', () => {
          store.nodeDelete(firstRootNode.id)
          expect(store.nodes[firstRootNode.id]).toBeUndefined()
        })
        it('should remove node id from roots list if node is root', () => {
          store.nodeDelete(firstRootNode.id)
          expect(store.referential.definition.roots.includes(firstRootNode.id)).toBe(false)
        })
        it('should remove all its descendants', () => {
          const descendants: string[] = firstRootNode.children.reduce((desc, childId) => {
            const childNode = baseReferential.definition.nodes[childId]
            desc.push(childId)
            if (childNode.children?.length) {
              desc.push(...childNode.children)
            }
            return desc
          }, [] as string[])
          store.nodeDelete(firstRootNode.id)
          descendants.forEach((descendantId) => {
            expect(store.referential.definition.nodes[descendantId]).toBeUndefined()
          })
        })
        it('should remove its reference from its parents children from', () => {
          const childNodeId = baseReferential.definition.nodes[firstRootNode.children[0]].id
          store.nodeDelete(childNodeId, true)
          expect((store.nodes[firstRootNode.id].children as string[]).includes(childNodeId)).toBe(false)
        })
        it('should not clean own reference in parents children, if review not asked', () => {
          const childNodeId = baseReferential.definition.nodes[firstRootNode.children[0]].id
          store.nodeDelete(childNodeId, false)
          expect((store.nodes[firstRootNode.id].children as string[]).includes(childNodeId)).toBe(true)
        })
      })
      describe('nodeEdit', () => {
        it('should update desired node', () => {
          const newNodeName = 'première racine'
          store.nodeEdit(firstRootNode.id, {name: newNodeName})
          expect(store.nodes[firstRootNode.id].name).toBe(newNodeName)
        })
      })
      describe('nodeMove', () => {
        describe('moving root node', () => {
          it('should move root from and to every position', () => {
            const newRoot = store.nodeCreate('root', 0)
            const rootCount = 1 + store.referential.definition.roots.length
            for (let i = 0; i < rootCount; i++) {
              for (let j = 0; j < rootCount; j++) {
                store.referential.definition.roots.splice(i, 0, newRoot.id)
                store.nodeMove(newRoot.id, i, j)
                const roots = store.referential.definition.roots
                if (i !== j) {
                  expect(roots[i] !== newRoot.id).toBe(true)
                }
                expect(roots[j]).toBe(newRoot.id)
                expect(roots.length).toEqual(rootCount)
                store.referential.definition.roots.splice(j, 1)
              }
            }
          })
        })
        describe('moving child node', () => {
          it('should move child from and to every position', () => {
            const newChild = store.nodeCreate('child', 0)
            const parent = store.nodes[firstRootId] as Required<Node>
            const childCount = 1 + parent.children.length
            for (let i = 0; i < childCount; i++) {
              for (let j = 0; j < childCount; j++) {
                store.referential.definition.nodes[parent.id].children?.splice(i, 0, newChild.id)
                store.nodeMove(newChild.id, i, j, parent.id)
                const children = store.referential.definition.nodes[parent.id].children as string[]
                if (i !== j) {
                  expect(children[i] !== newChild.id).toBe(true)
                }
                expect(children[j]).toBe(newChild.id)
                expect(children.length).toEqual(childCount)
                store.referential.definition.nodes[parent.id].children?.splice(j, 1)
              }
            }
          })
        })
      })
      describe('levelAdd', () => {
        const levelName = 'pays du monde'
        beforeEach(() => {
          store.levelAdd(levelName)
        })
        it('should add a level of given name', () => {
          const levels = store.referential.definition.levels
          expect(levels).toHaveLength(4)
          expect(levels[3].name).toEqual(levelName)
        })
        it('should have an uuid 4 as id', () => {
          const level = store.referential.definition.levels[3]
          expect(level.id.split('-')[2].startsWith('4')).toBe(true)
        })
      })
      describe('levelEdit', () => {
        it('should update desired level', () => {
          const newLevelName = 'premier niveau'
          store.levelEdit(firstLevel.id, {name: newLevelName})
          expect(store.levels[0].name).toBe(newLevelName)
        })
      })
      describe('levelDelete', () => {
        describe('root level deletion', () => {
          const rootLevelId = firstLevel.id
          it('should put next level nodes as roots', () => {
            const expectedRoots = Object.values(baseReferential.definition.nodes)
                .filter(node => node.level === 1)
            expectedRoots.sort((a, b) => a.name === b.name ? 0 : (a.name < b.name ? -1 : 1))

            store.levelDelete(rootLevelId)
            expect(store.referential.definition.roots)
                .toEqual(expectedRoots.map(node => node.id))
          })
          it('should delete orphan node of deeper level', () => {
            const roots = store.referential.definition.roots
            const nextOrphans = Object.values(store.nodes)
                .filter(node => node.level === 2 && roots.includes((node.parents as string[])[0]))
                .map(node => node.id)

            store.levelDelete(rootLevelId)
            const remainIds = Object.keys(store.nodes)
            nextOrphans.forEach(nodeId => {
              expect(remainIds.includes(nodeId)).toBe(false)
            })
          })
          it('should decrease all nodes level', () => {
            store.levelDelete(rootLevelId)
            const roots = store.rootNodes
            Object.values(store.nodes).forEach(node => {
              expect(node.level).toEqual(roots.includes(node) ? 0 : 1)
            })
          })
        })
        describe('last level deletion', () => {
          const lastLevelId = baseReferential.definition.levels[2].id
          it('should delete last level nodes', () => {
            store.levelDelete(lastLevelId)
            Object.values(store.nodes).forEach(node => {
              expect(node.level).toBeLessThan(2)
            })
          })
          it('should remove children reference from parents', () => {
            const baseNodes: Record<string, Node> = baseReferential.definition.nodes

            store.levelDelete(lastLevelId)
            const nodes = Object.values(store.nodes)
            nodes.forEach(node => {
              if (node.level === 0) {
                expect(node.children)
                    .toHaveLength((baseNodes[node.id].children as string[]).length - 1)
              }
              else {
                if (node.children) {
                  expect(node.children).toHaveLength(0)
                }
              }
            })

          })
          it('should keep level of each node', () => {
            const previousNodes: Record<string, Node> = baseReferential.definition.nodes
            store.levelDelete(lastLevelId)
            Object.values(store.nodes).forEach(node => {
              expect(node.level).toEqual(previousNodes[node.id].level)
            })
          })
        })
        describe('intermediate level', () => {
          const secondLevelId = baseReferential.definition.levels[1].id
          const secondLevelNodeIds = Object.values(baseReferential.definition.nodes)
              .reduce((ids: string[], node) => {
                if (node.level === 1) {
                  ids.push(node.id)
                }
                return ids
              }, [])
          it('should remove nodes from this level', () => {
            store.levelDelete(secondLevelId)
            secondLevelNodeIds.forEach(id => {
              expect(store.nodes[id]).toBeUndefined()
            })
          })
          it('should keep level consistency', () => {
            store.levelDelete(secondLevelId)
            Object.values(store.nodes).forEach(node => {
              if (node.level > 0) {
                expect(node.level).toEqual(baseReferential.definition.nodes[node.id].level - 1)
              }
            })
          })
          it('should remove node id from ancestor children list', () => {
            store.levelDelete(secondLevelId)
            Object.values(store.nodes).forEach(node => {
              if (node.children) {
                const noSecondLevelChild = node.children.every(id => !secondLevelNodeIds.includes(id))
                expect(noSecondLevelChild).toBe(true)
              }
            })
          })
          it('should add grandchildren to node parents children', () => {
            store.levelDelete(secondLevelId)
            secondLevelNodeIds.forEach((id) => {
              const {parents, children} = baseReferential.definition.nodes[id]
              if (children) {
                for (const ancestor of (parents as string[]).map(parentId => store.nodes[parentId])) {
                  expect(
                      children.every(childId => (ancestor.children as string[]).includes(childId))
                  ).toBe(true)
                }
              }
            })
          })
        })
      })
      describe('setDefaultLanguage', () => {
        it('should replace referential default language', () => {
          store.referential.languages = {
            available: ['fr', 'en'],
            default: 'fr',
            translations: {}
          }
          store.setDefaultLanguage('en')
          expect(store.defaultLanguage).toEqual('en')
        })
      })
      describe('updateTranslations', () => {
        it('should replace targeted node translations', () => {
          store.referential.languages = {
            available: ['fr'],
            default: 'fr',
            translations: {[firstRootId]: {'fr': 'première racine'}}
          }
          const replacementTranslation = {'fr': 'racine principale'}
          store.updateTranslations(firstRootId, replacementTranslation)
          expect(
              store.getTranslationFor(firstRootId, 'fr', 'error')
          ).toEqual(
              replacementTranslation['fr']
          )
        })
      })
    })
    describe('levelNodesLoad', () => {
      const storedNodes = baseReferential.definition.nodes
      const firstRootChildren: NodeBase[] = (storedNodes[firstRootId] as Required<Node>).children
          .map(nodeId => {
            const {id, name, parents} = storedNodes[nodeId]
            return {id, label: name, parents}
          })
      beforeEach(() => {
        levelNodesMock.mockResolvedValueOnce(firstRootChildren)

      })
      it('should call level node service', async () => {
        await store.levelNodesLoad(referentialId, secondLevel.id, 'fr')
        expect(levelNodesMock).toHaveBeenCalledOnce()
      })
      it('should initiate templateReferential entry in store', async () => {
        await store.levelNodesLoad(referentialId, secondLevel.id, 'fr')
        expect(Object.hasOwn(store.templateReferentials, referentialId)).toBe(true)
      })
      it('should replace templateReferential entry if exists in store', async () => {
        store.templateReferentials[referentialId] = {
          ...initReferential(),
          levelOptions: {
            [secondLevel.id]: [{label: 'random', value: 'uuid_v4', parents: []}]
          }
        }
        await store.levelNodesLoad(referentialId, secondLevel.id, 'fr')
        const secondLevelOptions = store.templateReferentials[referentialId].levelOptions[secondLevel.id]
        expect(secondLevelOptions).toHaveLength(firstRootChildren.length)
      })
    })
    describe('referentialCreate', () => {
      const projectId = 'da9f6ed4-eb2a-4b09-b48d-fae091e2cf17'
      const minimalRef: NewReferential = {
        name: 'new ref',
        description: 'describe me',
        scope: 4,
        languages: {available: ['fr'], default: 'fr', translations: {}},
        definition: {levels: [], roots: [], nodes: {}}
      }
      const createdRef: Referential = {
        ...baseReferential,
        ...minimalRef,
        project: {id: projectId, name: 'minimal project'}
      }
      beforeEach(async () => {
        postReferentialMock.mockResolvedValueOnce(createdRef)
        store.referential = minimalRef
        await store.referentialCreate(projectId)
      })
      it('should post project id and stored referential data to postReferential service', () => {
        expect(postReferentialMock).toHaveBeenNthCalledWith(1, {
          projectId,
          ...minimalRef
        })
      })
      it('should populate referential with received data', () => {
        expect(store.referential).toStrictEqual(createdRef)
      })
    })
    describe('referentialLoadForTemplate', () => {
      const receivedRef = initReferential()
      beforeEach(async () => {
        getReferentialMock.mockResolvedValueOnce(receivedRef)
        await store.referentialLoadForTemplate(referentialId, 'fr')
      })
      it('should get referential from getReferential service', () => {
        expect(getReferentialMock).toHaveBeenNthCalledWith(1, referentialId)
      })
      it('should populate referential with received data', () => {
        expect(store.referential).toStrictEqual(receivedRef)
      })
      it('should populate templateReferentials entry', () => {
        expect(Object.hasOwn(store.templateReferentials, referentialId))
        expect(store.templateReferentials[referentialId].id).toEqual(referentialId)
      })
      it('should populate templateReferentials levelOptions for each level', () => {
        const referentialEntry = store.templateReferentials[referentialId]
        baseReferential.definition.levels.forEach((level, i) => {
          expect(Object.hasOwn(referentialEntry.levelOptions, level.id)).toBe(true)
          // test some coherence in dispatch but not dispatch itself
          expect(referentialEntry.levelOptions[level.id].length).toEqual(Object.values(store.nodes).filter(n => n.level === i).length)
        })
      })
    })
    describe('referentialLoad', () => {
      const receivedRef = initReferential()
      beforeEach(async () => {
        getReferentialMock.mockResolvedValueOnce(receivedRef)
        await store.referentialLoad(referentialId)
      })
      it('should get referential from getReferential service', () => {
        expect(getReferentialMock).toHaveBeenNthCalledWith(1, referentialId)
      })
      it('should populate referential with received data', () => {
        expect(store.referential).toStrictEqual(receivedRef)
      })
    })
  })
})
